import os
from ModularCirc.Models.KorakianitisMixedModel import KorakianitisMixedModel, KorakianitisMixedModel_parameters, TEMPLATE_TIME_SETUP_DICT
from ModularCirc import BatchRunner
import numpy as np
import pandas as pd
from cvdnet_pipeline.utils import helper_functions, plot_utils

def simulate_data(param_path: str, 
                  n_samples: int, 
                  output_path: str, 
                  repeat_simulations: bool = True, 
                  sample_parameters: bool = True) -> str:
    
    # Sample parameters or read in from file
    
    br = BatchRunner('Sobol', 0)
    br.setup_sampler(param_path)
    br.sample(n_samples)

    if not sample_parameters:
        # print("Simulating from calibrated parameters.")
        posterior_samples = pd.read_csv(f"{output_path}/cleaned_posterior_samples.csv") 
        
        # remove any #s from column names
        posterior_samples.columns = posterior_samples.columns.str.lstrip('#').str.strip()

        # Align br._samples rows with those available in posterior_samples
        br._samples = br._samples.loc[posterior_samples.index].reset_index(drop=True)
       
        # Copy posterior values into br._samples
        for col in posterior_samples.columns:
            if col in br._samples.columns:
                br._samples.loc[:, col] = posterior_samples[col].values
            
    relevant_columns = [col for col in br._samples.columns if col in br._parameters_2_sample.keys()]
    _pure_samples = br._samples[relevant_columns].copy()

    
    map_ = {
            'delay': ['la.delay', 'ra.delay'],
            'td0': ['lv.td0', 'rv.td0'],
            'tr': ['lv.tr', 'rv.tr'],
            'tpww': ['la.tpww', 'ra.tpww'],
        }
        
    br.map_sample_timings(
        ref_time=1.,
        map=map_
    )

    br._samples[['lv.td', 'rv.td']] = br._samples[['lv.tr', 'rv.tr']].values + br._samples[['lv.td0', 'rv.td0']].values
    br._samples.drop(['lv.td0', 'rv.td0'], axis=1, inplace=True)

    # get relevant columns and count them
    n_params = len(relevant_columns)
    br.map_vessel_volume()

    br.setup_model(model=KorakianitisMixedModel, po=KorakianitisMixedModel_parameters,
                   time_setup=TEMPLATE_TIME_SETUP_DICT)

    file_suffix  = f'_{n_samples}_{n_params}_params'
    
    # Save as CSV files 
    _pure_samples.to_csv(os.path.join(output_path, f'pure_input{file_suffix}.csv'), index=False)
    br.samples.to_csv(os.path.join(output_path, f'input{file_suffix}.csv'), index=False)

    if sample_parameters:
        output_parameters = os.path.join(output_path, f'output{file_suffix}')
        output_parameters_simulations = os.path.join(output_parameters,'simulations')
        
    else:
        output_parameters = output_path
        output_parameters_simulations = os.path.join(output_parameters, f'posterior_simulations')

    # Check if the directory exists and contains n_samples files
    if os.path.exists(output_parameters_simulations) and len(os.listdir(output_parameters_simulations)) >= n_samples and repeat_simulations==False:
        print(f"Skipping simulation as {output_parameters} already contains {n_samples} or more files.")
        # read a list of dataframes from here
        simulations = helper_functions.load_simulation(output_parameters_simulations)

        if len(simulations) != n_samples:
            raise ValueError(f"Expected {n_samples} simulations, but found {len(simulations)}. Will run simulations again.")
            repeat_simulations = True
    else:
        print(f"Running simulation as {output_parameters}.")
        repeat_simulations = True


    if repeat_simulations:
        os.makedirs(output_parameters_simulations, exist_ok=True)
        simulations = br.run_batch(
            n_jobs=7,
            output_path=output_parameters_simulations
        )


    # Check for bool values in the list
    bool_indices = [index for index, value in enumerate(simulations) if isinstance(value, bool)]

    if bool_indices:
        print(f"Boolean values found at indices: {bool_indices}")
        print(f"Number of Booleans = {len(bool_indices)}")
    else:
        print("No boolean values found in the list.")


    helper_functions.save_csv(pd.DataFrame(bool_indices), os.path.join(output_parameters, f'bool_indices_{n_samples}.csv'))

    # plot simulated traces
    plot_utils.plot_simulated_traces(simulations, output_path=output_parameters)

    pressure_traces_df_pat, pressure_traces_df_rv = helper_functions.select_feasible_traces(simulated_traces=simulations, output_path=output_parameters)

    # Save the DataFrame to a single CSV file with headers
    helper_functions.save_csv(pressure_traces_df_pat, f'{output_parameters}/pressure_traces_pat/all_pressure_traces.csv')
    helper_functions.save_csv(pressure_traces_df_rv, f'{output_parameters}/pressure_traces_rv/all_pressure_traces.csv')

    plot_utils.plot_pressure_transients_arterial_tree(pressure_traces_df_rv, output_parameters)

    return output_parameters, n_params



from datetime import datetime
import numpy as np
import pandas as pd
from cvdnet_pipeline.utils.kf_emulator import KalmanFilterWithEmulator
from cvdnet_pipeline.utils.plot_utils import plot_kf_estimates
import os
import pickle

def KFGiessenSETUP(n_samples:int=4096, 
                n_params:int=9, 
                output_path:str='output', 
                emulator_path:str=None,
                output_keys:list=None,
                include_timeseries:bool=True,
                epsilon_obs_scale:float=0.05,
                ):
        
    # Load observation data
    output_file = pd.read_csv(f"{output_path}/waveform_resampled_all_pressure_traces_rv_with_pca.csv")

    # Input for priors
    input_prior = pd.read_csv(f'{emulator_path}/input_{n_samples}_{n_params}_params.csv')

    # emulators
    emulators = pd.read_pickle(f"{emulator_path}/output_{n_samples}_{n_params}_params/emulators/linear_models_and_r2_scores_{n_samples}.pkl")

    if include_timeseries:
        all_output_keys = output_file.iloc[:, :101].columns.tolist() + output_keys
        print("Including time-series in calibration as specified in config file.")

        # Build the diagonal entries: 101 ones followed by the std devs
        # 101 ones are scaled by epsilon_obs_scale so they will equal 
        # 1 when multipled by epsilon_obs_scale further down. 
        sd_values = output_file[output_keys].std().values
        diagonal_values = np.concatenate([np.ones(101)/epsilon_obs_scale, sd_values]) 
    else:
        all_output_keys = output_keys
        sd_values = output_file[output_keys].std().values
        diagonal_values = sd_values

    # Create the diagonal matrix
    e_obs = np.diag(diagonal_values) * epsilon_obs_scale
    
    # Select emulators and data for specified output_keys
    emulator_output = emulators.loc[all_output_keys]
    observation_data = output_file.loc[:, all_output_keys]

    # Priors
    mu_0 = np.array(input_prior.mean().loc[:'T'])
    mu_0 = mu_0.reshape(-1, 1)
    Sigma_0 = np.diag(input_prior.var().loc[:'T'])

    # dynamically define prior on T
    mu_0[-1,-1] = observation_data['iT'].iloc[0]
    Sigma_0[-1, -1] = 0.0000001

    # Parameter names
    param_names = input_prior.loc[:, :'T'].columns.to_list()

    # Model error
    epsilon_model = np.diag(emulator_output['RMSE']) 

    # Construct beta matrix and intercepts
    beta_matrix = []
    intercept = []

    for _, row_entry in emulator_output.iterrows():
        model = row_entry['Model']
        beta_matrix.append(model.coef_)
        intercept.append(model.intercept_)
    
    beta_matrix = np.array(beta_matrix)
    intercept = np.array(intercept).reshape(len(intercept), 1)
    
    # Process noise covariance
    Q = np.eye(n_params) * 0.01

    # Initialize the Kalman Filter with Emulator
    kf = KalmanFilterWithEmulator(beta_matrix, 
                                  intercept.flatten(), 
                                  Q, 
                                  e_obs, 
                                  epsilon_model, 
                                  mu_0.flatten(), 
                                  Sigma_0)

    # Run the filter
    estimates = kf.run(np.array(observation_data))

    # Save the resulting estimates

    # Define the output directory name, appending the number of output keys to the directory name and including a timestamp
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    output_dir_kf = f"{output_path}/kf_calibration_results/{len(all_output_keys)}_output_keys/calibration_{timestamp}"
    os.makedirs(output_dir_kf, exist_ok=True)

    # Save the estimated parameters to a CSV and npy files. First, turn the mu entries into a DataFrame
    mu_estimates_df = pd.DataFrame(
        np.array([estimate[0] for estimate in estimates]), 
        columns=param_names
    )
    sigma_estimates = np.array([estimate[1] for estimate in estimates])

    # Save to files
    mu_estimates_df.to_csv(f"{output_dir_kf}/kf_estimated_means.csv", index=False)
    np.save(f"{output_dir_kf}/kf_estimated_covariances.npy", sigma_estimates)

    # Save the entire estimates list as a pickle file
    with open(f"{output_dir_kf}/kf_estimated_means_and_covariances.pkl", 'wb') as f:
        pickle.dump(estimates, f)

    # Plot the results
    plot_kf_estimates(estimates=estimates, 
                      param_names=param_names,
                      output_path=output_dir_kf)

    return estimates


import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.stats import norm
import pandas as pd
import math


class BayesianCalibration:
    def __init__(self, 
                 input_prior, 
                 emulator_output, 
                 epsilon_obs, 
                 filtered_output=None, 
                 which_obs=None, 
                 observation_data=None, 
                 data_type="synthetic"):
        """
        Bayesian calibration for both synthetic and real data.
        
        For synthetic data: provide filtered_output and which_obs
        For real data: provide observation_data
        """
        self.input_prior = input_prior
        self.emulator_output = emulator_output
        self.epsilon_obs = epsilon_obs 
        self.data_type = data_type
        
        # Determine if we're using synthetic or real data
        if data_type == "real":
            # Real data mode
            self.observation_data = observation_data
        elif data_type == "synthetic":
            # Synthetic data mode
            self.filtered_output = filtered_output
            self.which_obs = which_obs
        else:
            raise ValueError("data_type must be either 'real' or 'synthetic'")

        # Setup priors
        self._setup_priors()
        
        # Parameter names
        self.param_names = input_prior.columns.to_list() # Unclear if need to do this for real data. The .loc stuff.

        # Model error
        self.epsilon_model = np.diag(emulator_output['RMSE']) 
        
        # Compute posterior
        self.compute_posterior()


    def _setup_priors(self):
        """Setup prior parameters based on data type"""

        self.mu_0 = np.array(self.input_prior.mean())
        self.ind = self.input_prior.columns.get_loc("T")

        if self.data_type == "synthetic":
            # dynamically define prior on T
            self.mu_0[self.ind] = self.filtered_output['iT'].iloc[self.which_obs]
            self.mu_0 = self.mu_0.reshape(-1, 1)
            self.Sigma_0 = np.diag(self.input_prior.var())

        elif self.data_type == "real":

            self.mu_0 = self.mu_0.reshape(-1, 1)
            self.Sigma_0 = np.diag(self.input_prior.var())

            # dynamically define prior on T
            self.mu_0[self.ind,-1] = self.observation_data['iT'].iloc[0]

        self.Sigma_0[self.ind, self.ind] = 0.01


    def compute_posterior(self):
        full_error = self.epsilon_obs + self.epsilon_model
       
        # Construct beta matrix and intercepts
        beta_matrix = []
        intercept = []

        for _, row_entry in self.emulator_output.iterrows():
            model = row_entry['Model']
            beta_matrix.append(model.coef_)
            intercept.append(model.intercept_)
        
        beta_matrix = np.array(beta_matrix)
        intercept = np.array(intercept).reshape(len(intercept), 1)

        # Select observation based on data type
        if self.data_type == "real":
            Y_obs = np.array(self.observation_data.T).reshape(-1, 1)
        else:
            Y_obs = np.array(self.filtered_output.T[self.which_obs]).reshape(-1, 1)
        
        Y_scaled = Y_obs - intercept

        # Compute posterior covariance
        Sigma_post_inv = (beta_matrix.T @ np.linalg.inv(full_error) @ beta_matrix) + np.linalg.inv(self.Sigma_0)
        self.Sigma_post = np.linalg.inv(Sigma_post_inv)
        
        # Compute posterior mean
        self.Mu_post = self.Sigma_post @ (beta_matrix.T @ np.linalg.inv(full_error) @ Y_scaled + np.linalg.inv(self.Sigma_0) @ self.mu_0)


    def sample_posterior(self, n_samples):
        rg = np.random.default_rng(1)
        self.samples = rg.multivariate_normal(self.Mu_post.flatten(), self.Sigma_post, size=n_samples)
        self.samples_df = pd.DataFrame(self.samples)
        self.samples_df.columns = self.param_names


import pytest
import pandas as pd
import os
from cvdnet_pipeline.analyse_giessen import analyse_giessen


# Parametrize data types for testing both synthetic and real data
DATA_TYPES = ["synthetic","real"]


@pytest.fixture(params=DATA_TYPES)
def data_type(request):
    """Fixture that provides both synthetic and real data types for parameterized tests."""
    return request.param


@pytest.fixture
def cleanup_output_file(data_type):
    """Cleanup fixture that handles output files for both data types."""

    if data_type == "real":
        output_file = f"tests/inputs_for_tests/analyse_giessen_module/output_64_9_params/{data_type}_data/pressure_traces_rv/waveform_resampled_all_pressure_traces_rv.csv"
    elif data_type == "synthetic":
        output_file = f"tests/inputs_for_tests/analyse_giessen_module/output_64_9_params/{data_type}_data/waveform_resampled_all_pressure_traces_rv.csv"
    yield output_file, data_type
    if os.path.exists(output_file):
        os.remove(output_file)

@pytest.fixture
def cleanup_calibration_output_file():
    """Cleanup fixture for calibrated synthetic data (only available for synthetic data currently)."""
    output_file = "tests/inputs_for_tests/analyse_giessen_module/output_64_9_params/synthetic_data/bayesian_calibration_results/17_output_keys/calibration_20250604_154542/waveform_resampled_all_pressure_traces_rv.csv"
    yield output_file
    if os.path.exists(output_file):
        os.remove(output_file)        

def test_analyse_giessen_valid_input(cleanup_output_file):
    """Test analyse_giessen with valid input for both synthetic and real data."""
    output_file, data_type = cleanup_output_file
    
    filepath = f'tests/inputs_for_tests/analyse_giessen_module/output_64_9_params/{data_type}_data/'

    if data_type == "real":
        filepath += "pressure_traces_rv/"

    # Call the function with the input file
    analyse_giessen(file_path=filepath, 
                    data_type=data_type,
                    gaussian_sigmas=[6., 4., 2.])

    # Check if the output data matches the expected output
    output_data = pd.read_csv(output_file)
    if data_type == "real":
        expected_output = pd.read_csv(f'tests/expected_outputs/analyse_giessen_module/output_64_9_params/{data_type}_data/pressure_traces_rv/waveform_resampled_all_pressure_traces_rv.csv')
    elif data_type == "synthetic":
        expected_output = pd.read_csv(f'tests/expected_outputs/analyse_giessen_module/output_64_9_params/{data_type}_data/waveform_resampled_all_pressure_traces_rv.csv')   
    pd.testing.assert_frame_equal(output_data[expected_output.columns], expected_output)


@pytest.mark.parametrize("data_type", DATA_TYPES)
def test_analyse_giessen_invalid_input(data_type):
    """Test analyse_giessen with invalid file path for both data types."""
    # Test with an invalid file path
    with pytest.raises(FileNotFoundError):
        analyse_giessen(file_path="invalid/path",
                        data_type=data_type,
                        gaussian_sigmas=[6., 4., 2.])

def test_analyse_giessen_valid_calibrated_input(cleanup_calibration_output_file):
    """Test analyse_giessen with valid calibrated input (synthetic data only currently)."""

    # Call the function with the input file
    analyse_giessen(file_path='tests/inputs_for_tests/analyse_giessen_module/output_64_9_params/synthetic_data/bayesian_calibration_results/17_output_keys/calibration_20250604_154542/',
                    data_type="synthetic",
                     gaussian_sigmas=[6., 4., 2.])

    # Check if the output data matches the expected output
    output_data = pd.read_csv(cleanup_calibration_output_file)
    expected_output = pd.read_csv('tests/expected_outputs/analyse_giessen_module/output_64_9_params/synthetic_data/bayesian_calibration_results/17_output_keys/calibration_20250604_154542/waveform_resampled_all_pressure_traces_rv.csv')
    pd.testing.assert_frame_equal(output_data[expected_output.columns], expected_output)



import os
import pytest
import pandas as pd
from cvdnet_pipeline.calibrate_parameters import calibrate_parameters
import tempfile
import shutil
import glob
from tests.test_constants import (
    OUTPUT_KEYS_FOR_TESTS_FOR_REAL_DATA_CALIBRATION, 
    DEFAULT_N_SAMPLES, 
    DEFAULT_N_PARAMS, 
    DEFAULT_EPSILON_OBS_SCALE
)


def test_calibrate_parameters_real():
    # Define test parameters

    with tempfile.TemporaryDirectory() as tmp_path:
        print(f"Temporary directory created at: {tmp_path}")

        output_keys = OUTPUT_KEYS_FOR_TESTS_FOR_REAL_DATA_CALIBRATION
        n_samples = DEFAULT_N_SAMPLES
        n_params = DEFAULT_N_PARAMS

        # Copy all the expected input files to the temporary directory
        emulator_path = './tests/known_good_outputs/synthetic_data/'
        # From emulator_path, copy the input file: pure_input_64_9_params.csv
        shutil.copy(os.path.join(emulator_path, f'pure_input_{n_samples}_{n_params}_params.csv'), tmp_path)
        # Now copy over the waveform data:
        shutil.copy('./tests/known_good_outputs/real_data/waveform_resampled_all_pressure_traces_rv_with_pca.csv', tmp_path)

        calibrate_parameters(data_type="real",
                             n_samples=n_samples,
                             n_params=n_params,
                             output_path=str(tmp_path),
                             emulator_path=emulator_path,
                             output_keys=output_keys,
                             include_timeseries=False,
                             epsilon_obs_scale=DEFAULT_EPSILON_OBS_SCALE,
                             config=[])

        # Compare the output files to the expected output files

        # Load the expected output ----------------------------------------------------------------------
        expected_output_dir = os.path.join(
            './tests/known_good_outputs/real_data/',
            'bayesian_calibration_results',
            '15_output_keys',
            'calibration_20250911_133944'
        )
        
        expected_posterior_covariance = pd.read_csv(os.path.join(expected_output_dir,
                                                    'posterior_covariance.csv'))
        expected_posterior_mean       = pd.read_csv(os.path.join(expected_output_dir,
                                                    'posterior_mean.csv'))                
        
        # List all files in the tmp_path to verify the output files
        print("Files in the temporary directory:")
        for root, dirs, files in os.walk(tmp_path):
            for file in files:
                print(os.path.join(root, file))


        # Load the actual output files ------------------------------------------------------------------

        # Find the actual calibration_* directory, because the name is created dynamically with a timestamp
        calibration_dirs = glob.glob(os.path.join(
            tmp_path,
            'bayesian_calibration_results/15_output_keys',
            'calibration_*'
        ))

        print(f"Found calibration directories: {calibration_dirs}")
        assert len(calibration_dirs) == 1, "Expected exactly one calibration_* directory"
        calibration_dir = calibration_dirs[0]

        posterior_covariance = pd.read_csv(os.path.join(calibration_dir,
                                                        'posterior_covariance.csv'))
        posterior_mean       = pd.read_csv(os.path.join(calibration_dir,
                                                        'posterior_mean.csv'))              

        # Test rough equality ------------------------------------------------------------------------
        pd.testing.assert_frame_equal(
            expected_posterior_covariance, 
            posterior_covariance, 
            atol=1e-4, 
            rtol=1e-4, 
            check_exact=False
        )
        pd.testing.assert_frame_equal(
            expected_posterior_mean, 
            posterior_mean, 
            atol=1e-4, 
            rtol=1e-4, 
            check_exact=False
        )




import json
from cvdnet_pipeline.simulate_data import simulate_data
from cvdnet_pipeline.analyse_giessen import analyse_giessen
from cvdnet_pipeline.compute_pca import compute_pca
from cvdnet_pipeline.build_emulator import build_emulator
from cvdnet_pipeline.calibrate_parameters import calibrate_parameters
from cvdnet_pipeline.sensitivity_analysis import sensitivity_analysis
from cvdnet_pipeline.utils import plot_utils
from cvdnet_pipeline.kalman_filter_giessen import KFGiessenSETUP
import os
import argparse
from cvdnet_pipeline.utils.constants import VALID_PIPELINE_STEPS

def run_pipeline(config):

    steps = config.get("steps", VALID_PIPELINE_STEPS)

    # If steps has invalid entries, raise an error
    for step in steps:
        if step not in VALID_PIPELINE_STEPS:
            raise ValueError(f"Invalid step '{step}' found in steps. "
                             f"Valid steps for the pipeline are: {VALID_PIPELINE_STEPS}")

    data_type = config.get("data_type", "synthetic")

    if data_type == "synthetic":

        print("Processing the pipeline for synthetic data.")

        n_samples = config.get("n_samples", 2048)

        # Parent folder for all simulations
        output_path = config.get("output_path")

        if not "sim" in steps:
            # Get the n_params from the config if not provided by step 1
            n_params = config.get("n_params")

            if n_params is None:
                raise ValueError("n_params must be provided in the configuration if step 1 is not being executed.")

            # Define the output directory for the current simulations
            output_dir_sims = os.path.join(output_path, f'output_{n_samples}_{n_params}_params')
            print("Simulation output directory is: ", output_dir_sims)

        os.makedirs(output_path, exist_ok=True)

        if "sim" in steps:
            print("Step 1: Simulating Data")

            if "n_params" in locals():
                print("Warning: n_params is pre-defined in the configuration file. It will be overwritten by the value from the simulation step.")

            if "output_dir_sims" in locals():
                print("Warning: output_dir_sims is pre-defined in the configuration file. It will be overwritten by the value from the simulation step.")

            output_dir_sims, n_params = simulate_data(
                param_path=os.path.join('./input_parameters_jsons', config.get("input_parameters")),
                n_samples=n_samples,
                output_path=output_path,
                sample_parameters=True
            )

        if "ag" in steps:
            print("Step 2: Analysing Giessen (resample)")
            analyse_giessen(file_path=output_dir_sims,
                            data_type=data_type,
                            gaussian_sigmas=config.get('gaussian_sigmas'),
            )

        if "pca" in steps:
            print("Step 3: Compute PCA")

            n_pca_components = config.get("n_pca_components", 10)
            if n_pca_components is None:
                raise ValueError("n_pca_components must be provided in the configuration to run PCA.")

            compute_pca(n_samples=n_samples, 
                        n_params=n_params, 
                        n_pca_components=n_pca_components,
                        output_path=output_path,
                        data_type=data_type)

        if "emu" in steps:
            print("Step 4: Building Emulator")
            output_keys = config.get("output_keys", None)
            build_emulator(n_samples=n_samples,
                        n_params=n_params, 
                        output_path=output_path, 
                        output_file_name="waveform_resampled_all_pressure_traces_rv_with_pca.csv",
                        output_keys_red=output_keys)

        if "gsa" in steps:
            print("Step GSA: Global Sensitivity Analysis")
            sensitivity_analysis(n_samples=n_samples,
                                 n_params=n_params, 
                                 output_path=output_path,
                                 n_processes=None)

        if "cal" in steps:
            print("Step 5: Calibrating parameters using config output keys")

            output_keys = config.get("output_keys")
            if output_keys is None:
                raise ValueError("output keys must be provided in the configuration to run calibration.")
            
            include_timeseries = bool(config.get("include_timeseries"))
            
            dummy_data_dir = config.get("dummy_data_dir")
            if dummy_data_dir is None and not data_type == "real":
                raise ValueError("Dummy data directory 'dummy_data_dir', must be provided in the configuration to run calibration for synthetic data.")
            
            output_dir_bayesian, e_obs = calibrate_parameters(
                                        data_type=data_type,
                                        n_samples=n_samples,
                                        n_params=n_params,
                                        output_path=output_path,
                                        output_keys=output_keys,
                                        include_timeseries=include_timeseries,
                                        epsilon_obs_scale=config.get("epsilon_obs_scale", 0.05),
                                        dummy_data_dir=dummy_data_dir,
                                        config=config)

        if "post_sim" in steps:
            print("Step 6: Simulating posterior pressure waves.")
        
        
            if not "cal" in steps:
                output_dir_bayesian = config.get("output_dir_bayesian")
                print(f"Reading parameter file from {output_dir_bayesian} as pre-defined in the configuration file.")

            output_dir_bayesian, n_params = simulate_data(
                param_path=os.path.join('./input_parameters_jsons', config.get("input_parameters")),
                n_samples=n_samples,
                output_path=output_dir_bayesian,
                sample_parameters = False
            )

        if "post_res" in steps:
            print("Step 7: Resampling posterior pressure waves.")

            if not "post_sim" in steps:
                output_dir_bayesian = config.get("output_dir_bayesian")
                print(f"Loading posterior samples from {output_dir_bayesian} as pre-defined in the configuration file.")

            analyse_giessen(file_path=output_dir_bayesian, 
                            data_type=data_type,
                            gaussian_sigmas=config.get('gaussian_sigmas')
                            )
            dummy_data_dir = config.get("dummy_data_dir")
            print(f"Reading true waveform from {dummy_data_dir} as pre-defined in the configuration file.")
            plot_utils.plot_posterior_simulations(dummy_data_dir, output_dir_bayesian)

        print("Pipeline complete.")

    elif data_type == "real":
        print("Processing the pipeline for real data.")
        
        # Parent folder for all simulations
        output_path = config.get("output_path")
        
        # Define the output directory for the current data
        print("Data directory is: ", output_path)
        
        if "ag" in steps:
            print("Step 2: Analysing Giessen (resample)")
            analyse_giessen(file_path=output_path,
                            data_type=data_type,
                            gaussian_sigmas=config.get('gaussian_sigmas'))

        if "pca" in steps:
            print("Step 3: Compute PCA")

            n_pca_components = config.get("n_pca_components", 10)
            if n_pca_components is None:
                raise ValueError("n_pca_components must be provided in the configuration to run PCA.")

            compute_pca(n_pca_components=n_pca_components,
                        output_path=output_path,
                        data_type=data_type,)

        if "cal" in steps:
            print("Step 5: Calibrating parameters using config output keys")

            output_keys = config.get("output_keys")
            if output_keys is None:
                raise ValueError("output keys must be provided in the configuration to run calibration.")
            
            emulator_path = config.get("emulator_path")
            n_samples = config.get("n_samples")
            n_params = config.get("n_params")
            include_timeseries = bool(config.get("include_timeseries"))

            dummy_data_dir = config.get("dummy_data_dir")
            if dummy_data_dir is None and not data_type == "real":
                raise ValueError("Dummy data directory 'dummy_data_dir', must be provided in the configuration to run calibration for synthetic data.")

            calibrate_parameters(data_type=data_type,
                                 n_samples=n_samples,
                                 n_params=n_params,
                                 output_path=output_path,
                                 emulator_path=emulator_path,
                                 output_keys=output_keys,
                                 include_timeseries=include_timeseries,
                                 dummy_data_dir=dummy_data_dir,
                                 config=config)

        if "kf" in steps:
            print("Step 5: Kalman Filter with Emulator")

            output_keys = config.get("output_keys")
            if output_keys is None:
                raise ValueError("output keys must be provided in the configuration to run calibration.")
            
            emulator_path = config.get("emulator_path")
            n_samples = config.get("n_samples")
            n_params = config.get("n_params")
            include_timeseries = bool(config.get("include_timeseries"))        

            estimates  = KFGiessenSETUP(n_samples=n_samples,
                n_params=n_params,
                output_path=output_path,
                emulator_path=emulator_path,
                output_keys=output_keys,
                include_timeseries=include_timeseries,
                epsilon_obs_scale=0.05)    
        
        print("Pipeline complete.")
        
    else:
        raise ValueError(f"Unknown data type: {data_type}. Supported types are 'synthetic' and 'real'.")

def main():
    """
    CLI entry point for running the CVDNet pipeline.

    This function parses command-line arguments to obtain the path to a JSON configuration file,
    loads the configuration, and executes the pipeline accordingly.

    Usage:
        python -m cvdnet_pipeline.run_pipeline --config path/to/config.json
    """
    parser = argparse.ArgumentParser(description="Run the pipeline with a configuration file.")
    parser.add_argument(
        "--config",
        type=str,
        required=True,
        help="Path to the configuration file (JSON format)."
    )
    args = parser.parse_args()

    # Load configuration from the specified JSON file
    with open(args.config, "r") as config_file:
        config = json.load(config_file)

    run_pipeline(config)

if __name__ == "__main__":
    main()



import pandas as pd
import numpy as np
from cvdnet_pipeline.utils.bayesian_calibration import BayesianCalibration
import os
from cvdnet_pipeline.utils import plot_utils
import json
from datetime import datetime

def calibrate_parameters(data_type="synthetic",
                         n_samples:int=64, 
                         n_params:int=9, 
                         output_path:str='output', 
                         emulator_path:str=None,
                         output_keys:list=None,
                         include_timeseries:bool=True,
                         epsilon_obs_scale:float=0.05,
                         dummy_data_dir:str=None,
                         config:dict=None):

    if data_type == "synthetic":
    
        file_suffix = f'_{n_samples}_{n_params}_params'

        dir_name = f"{output_path}/output{file_suffix}"

        input_params = pd.read_csv(f'{output_path}/pure_input{file_suffix}.csv')
        
        # True input parameters of dummy data
        true_input = pd.read_csv(f"{dummy_data_dir}/input_dummy_data.csv")

        # Synthetic dummy data (to calibrate on)
        output_file = pd.read_csv(f"{dummy_data_dir}/output_dummy_data/waveform_resampled_all_pressure_traces_rv_with_pca.csv")

        # emulators
        emulators = pd.read_pickle(f"{dir_name}/emulators/linear_models_and_r2_scores_{n_samples}.pkl")
        
    elif data_type == "real":

        dir_name = output_path

        # Real data (to calibrate on)
        output_file = pd.read_csv(f"{dir_name}/waveform_resampled_all_pressure_traces_rv_with_pca.csv")

        input_params = pd.read_csv(f'{emulator_path}/pure_input_{n_samples}_{n_params}_params.csv')

        # emulators
        emulators = pd.read_pickle(f"{emulator_path}/output_{n_samples}_{n_params}_params/emulators/linear_models_and_r2_scores_{n_samples}.pkl")
        print(f"Using trained emulators from: {emulator_path}/output_{n_samples}_{n_params}_params.")

    # Direcotry for saving results
    output_dir = f"{dir_name}/bayesian_calibration_results/"

    # Make directory if it doesn't exist
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    
    if include_timeseries:
        all_output_keys = output_file.iloc[:, :101].columns.tolist() + output_keys
        print("Including time-series in calibraiton as specified in config file.")

        # Build the diagonal entries: 101 ones followed by the std devs
        # 101 ones are scaled by epsilon_obs_scale so they will equal 
        # 1 when multipled by epsilon_obs_scale further down. 
        sd_values = output_file[output_keys].std().values
        diagonal_values = np.concatenate([np.ones(101)/epsilon_obs_scale, sd_values]) 
    else:
        all_output_keys = output_keys
        sd_values = output_file[output_keys].std().values
        diagonal_values = sd_values


    # Select emulators and data for specified output_keys
    emulator_output = emulators.loc[all_output_keys]
    observation_data = output_file.loc[:, all_output_keys] 
    
    if data_type == "synthetic":

        # Create the diagonal matrix
        e_obs = np.diag(diagonal_values) * epsilon_obs_scale
        
        bc = BayesianCalibration(input_prior=input_params, 
                                emulator_output=emulator_output, 
                                filtered_output=observation_data, 
                                which_obs=3, 
                                epsilon_obs = e_obs,
                                )

        bc.compute_posterior()

        # Save the posterior mean and covariance
        posterior_mean = pd.DataFrame(bc.Mu_post, index=bc.param_names, columns=['Posterior Mean'])    
        posterior_cov = pd.DataFrame(bc.Sigma_post, index=bc.param_names, columns=bc.param_names)

        
        # Sample from the posterior distribution
        bc.sample_posterior(n_samples=n_samples)

    elif data_type == "real":

        posterior_means = []

        # Create the diagonal matrix
        e_obs = np.diag(diagonal_values) * epsilon_obs_scale

        for row in range(len(observation_data)):
            bc = BayesianCalibration(input_prior=input_params, 
                                     emulator_output=emulator_output, 
                                     observation_data=observation_data.iloc[row:row+1], 
                                     epsilon_obs = e_obs,
                                     data_type=data_type)
            bc.compute_posterior()
            posterior_means.append(bc.Mu_post.squeeze())

        # Convert the list to a NumPy array
        posterior_means = np.array(posterior_means)
        Sigma_post = bc.Sigma_post

        # Save the posterior mean and covariance
        posterior_mean = pd.DataFrame(posterior_means, columns=bc.param_names)    
        posterior_cov = pd.DataFrame(Sigma_post, index=bc.param_names, columns=bc.param_names)

    n_output_keys =  len(all_output_keys)

    # Define the output directory name, appending the number of output keys to the directory name and including a timestamp
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    output_dir_bayesian = f"{output_dir}/{n_output_keys}_output_keys/calibration_{timestamp}"

    # Check if the directory exists, if not, create it
    if not os.path.exists(output_dir_bayesian):
        os.makedirs(output_dir_bayesian)

    posterior_mean.to_csv(f"{output_dir_bayesian}/posterior_mean.csv", index=False)
    posterior_cov.to_csv(f"{output_dir_bayesian}/posterior_covariance.csv", index=False)

    if data_type == "synthetic":
        bc.samples_df.to_csv(f"{output_dir_bayesian}/posterior_samples.csv", index=False)
        
        # Remove negative samples
        cleaned_samples = bc.samples_df[(bc.samples_df >= 0).all(axis=1)]

         # Flag number of posterior samples that were removed during cleaning
        missing_indices = set(bc.samples_df.index) - set(cleaned_samples.index)
        if missing_indices:
            print(f"The following posterior_samples were removed due to negativity: {sorted(missing_indices)}")
            pd.Series(sorted(missing_indices)).to_csv(f"{output_dir_bayesian}/cleaned_posterior_samples_indices.csv", index=False)
       
        cleaned_samples.to_csv(f"{output_dir_bayesian}/cleaned_posterior_samples.csv", index=False)

    # Save the config file
    with open(os.path.join(output_dir_bayesian, 'used_config.json'), 'w') as f:
        json.dump(config, f, indent=4)

    if data_type == "synthetic":
        
        # Plot the prior and posteior distributions
        plot_utils.plot_posterior_distributions(true_input, 
                                                bc.mu_0,
                                                bc.Sigma_0,
                                                bc.Mu_post,
                                                bc.Sigma_post,
                                                bc.which_obs,
                                                bc.param_names,
                                                output_path=output_dir_bayesian)

        # Plot posterior covariance matrix
        plot_utils.plot_posterior_covariance_matrix(bc.Sigma_0,
                                                    bc.Sigma_post,
                                                    bc.param_names,
                                                    output_path=output_dir_bayesian)
        
    elif data_type == "real":

        plot_utils.plot_parameter_trajectories(Sigma_post=Sigma_post,
                                               posterior_means=posterior_means,
                                               bc=bc,
                                               output_path=output_dir_bayesian)

    
    return output_dir_bayesian, e_obs


import matplotlib.pyplot as plt
import scipy.stats as stats
import matplotlib.cm as cm

import numpy as np
import os
import pandas as pd
from sklearn.pipeline import Pipeline

import seaborn as sns
from scipy.stats import norm

import matplotlib.pyplot as plt

import numpy as np
import math
import os


def plot_simulated_traces(simulated_traces, output_path):
    def save_plot(x, y, xlabel, ylabel, title, filename):
        """Helper function to create and save a plot."""
        fig, ax = plt.subplots()
        for indices in range(len(simulated_traces)):
            if not isinstance(simulated_traces[indices], bool):
                ax.plot(x(indices), y(indices), label=f'Realisation {indices}')
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title(title)
        fig.tight_layout()
        plt.savefig(filename)
        plt.close(fig)

    # Create output directory for figures
    output_path_figures = os.path.join(output_path, "figures")
    os.makedirs(output_path_figures, exist_ok=True)

    # Plot 1: Pressure Transients in Arterial Tree
    save_plot(
        x=lambda i: simulated_traces[i]['T'] - simulated_traces[i]['T'].iloc[0],
        y=lambda i: simulated_traces[i]['p_pat'],
        xlabel='Time (seconds)',
        ylabel='Pressure (mmHg)',
        title='Pressure Transients in Arterial Tree',
        filename=os.path.join(output_path_figures, 'pressure_transients_arterial_tree_1.png')
    )

    # Plot 2: Resampled Pressure Transients in Arterial Tree
    save_plot(
        x=lambda i: list(range(100)),
        y=lambda i: np.interp(
            np.linspace(
                simulated_traces[i]['T'].iloc[0],
                simulated_traces[i]['T'].iloc[-1],
                100
            ),
            simulated_traces[i]['T'],
            simulated_traces[i]['p_pat']
        ),
        xlabel='Time index',
        ylabel='Pressure (mmHg)',
        title='Resampled Pressure Transients in Arterial Tree',
        filename=os.path.join(output_path_figures, 'pressure_transients_arterial_tree_100.png')
    )

    # Plot 3: Resampled Pressure Transients in RV
    save_plot(
        x=lambda i: list(range(100)),
        y=lambda i: np.interp(
            np.linspace(
                simulated_traces[i]['T'].iloc[0],
                simulated_traces[i]['T'].iloc[-1],
                100
            ),
            simulated_traces[i]['T'],
            simulated_traces[i]['p_rv']
        ),
        xlabel='Time index',
        ylabel='Pressure (mmHg)',
        title='Resampled Pressure Transients in RV',
        filename=os.path.join(output_path_figures, 'pressure_transients_RV.png')
    )


def plot_pressure_transients_arterial_tree(input_traces, output_path):
    """
    Plot pressure transients in the arterial tree and save the figures.

    Parameters:
        simulated_traces (list): List of simulated traces.
        output_path (str): Path to save the figures.
    """

    output_path_figures = os.path.join(output_path,"figures")
    os.makedirs(output_path_figures, exist_ok=True)

    fig, ax = plt.subplots()

    for ind in range(len(input_traces)):
        t = range(100)  # Time adjustment
        p_pat = input_traces.iloc[ind, :100].values  # Pressure transient

        # Plot the pressure transient for each realization
        ax.plot(t, p_pat, label=f'Realisation {ind}')

    # Set labels and title
    ax.set_xlabel('Time Index')
    ax.set_ylabel('Pressure (mmHg)')
    ax.set_title('Pressure Transients in Arterial Tree')

    fig.tight_layout()

    # Display the plot
    plt.savefig(f'{output_path_figures}/pressure_transients_arterial_tree_good_traces.png')


def plot_pca_explained_variance(pipeline, output_path):
    pipeline : Pipeline
    pca = pipeline.named_steps['pca']

    output_path_figures = os.path.join(output_path,"figures")
    os.makedirs(output_path_figures, exist_ok=True)

    # Create figure
    fig, axs = plt.subplots(1, 2)
    n = pca.n_components_
    grid = np.arange(1, n + 1)
    # Explained variance
    explained_variance_ratio = pca.explained_variance_ratio_
    axs[0].bar(grid, explained_variance_ratio, log=True)
    axs[0].set(xlabel="Component", title="% Explained Variance", ylim=(1e-3, 1.0))

    # Cumulative Variance
    cumulative_explained_variance = np.cumsum(explained_variance_ratio)
    axs[1].semilogy(grid, cumulative_explained_variance, "o-")
    axs[1].set(
        xlabel="Component", title="% Cumulative Variance", 
    )
    # Set up figure
    fig.set(figwidth=8, dpi=100)
    fig.tight_layout()
            
    plt.savefig(f'{output_path_figures}/pca_explained_variance.png')


def plot_pca_transformed(pipeline, X, output_path):
    pipeline : Pipeline

    output_path_figures = os.path.join(output_path,"figures")
    os.makedirs(output_path_figures, exist_ok=True)

    signals_pca = pipeline.transform(X)

    fig, ax = plt.subplots(ncols=10, nrows=2, figsize=(70, 15))

    for i in range(signals_pca.shape[1]):
        temp = np.zeros(signals_pca.shape)
        temp[:, i] = signals_pca[:, i]
        
        signals_new = pipeline.inverse_transform(temp)
        
        ax[1][i].hist(signals_pca[:,i], bins=10)
        for signal in signals_new:
            ax[0][i].plot(signal)
    
    plt.savefig(f'{output_path_figures}/pca_transformed.png')
            

def plot_pca_histogram(X_pca, output_path, n_pca_components=10):


    output_path_figures = os.path.join(output_path,"figures")

    try:
        X_pca.hist(figsize=(15, 13), layout=(5, 2), alpha=0.7, color='orange', bins=30)
    except Exception:
        X_pca.hist(figsize=(15, 13), layout=(5, 2), alpha=0.7, color='orange')
    
    plt.tight_layout()
    plt.suptitle(f'Histograms of the First {n_pca_components} Principal Components')
    plt.savefig(f'{output_path_figures}/histograms_pca.png')    


def plot_posterior_distributions(true_input, mu_0, Sigma_0, Mu_post, Sigma_post, which_obs, param_names, output_path):

    output_path_figures = os.path.join(output_path,"figures")
    os.makedirs(output_path_figures, exist_ok=True)

    prior_means = mu_0.flatten()
    prior_stds = np.sqrt(np.diag(Sigma_0))
    posterior_means = Mu_post.flatten()
    posterior_stds = np.sqrt(np.diag(Sigma_post))
    true_values = true_input.loc[which_obs, param_names].values
   
    fig, axes = plt.subplots(2, math.ceil(len(param_names)/2), figsize=(18, 8))  
    axes = axes.flatten()  # Flatten to 1D array
    for i, ax in enumerate(axes[:len(param_names)]):  # Only iterate over valid axes
        # Define x-range based on prior and posterior means
        x_min = min(prior_means[i] - 3 * prior_stds[i], posterior_means[i] - 3 * posterior_stds[i])
        x_max = max(prior_means[i] + 3 * prior_stds[i], posterior_means[i] + 3 * posterior_stds[i])
        x = np.linspace(x_min, x_max, 100)

        # Compute PDFs
        prior_pdf = norm.pdf(x, prior_means[i], prior_stds[i])
        posterior_pdf = norm.pdf(x, posterior_means[i], posterior_stds[i])

        # Plot prior and posterior distributions
        ax.plot(x, prior_pdf, label="Prior", linestyle="dashed", color="blue")
        ax.plot(x, posterior_pdf, label="Posterior", linestyle="solid", color="red")

        # Plot true value as a vertical line
        ax.axvline(true_values[i], color="green", linestyle="dotted", label="True Value")

        # Labels and title
        ax.set_title(param_names[i])
        ax.set_xlabel("Value")
        ax.set_ylabel("Density")
        # Remove per-axis legends
        # ax.legend()

    # Remove unused axes if any
    for ax in axes[len(param_names):]:
        ax.remove()


    fig.tight_layout()
    plt.subplots_adjust(bottom=0.14)

    # Add a single legend across the bottom
    handles = [
        plt.Line2D([0], [0], color="blue",  linestyle="dashed", label="Prior"),
        plt.Line2D([0], [0], color="red",   linestyle="solid",  label="Posterior"),
        plt.Line2D([0], [0], color="green", linestyle="dotted", label="True Value")
    ]
    
    fig.legend(handles=handles, ncol=3, fontsize=14, loc="lower center")
    plt.savefig(f'{output_path_figures}/posterior_distributions_calibrated_params.png')    
    

def plot_posterior_covariance_matrix(Sigma_0, Sigma_post, param_names, output_path):
        
    output_path_figures = os.path.join(output_path,"figures")
    os.makedirs(output_path_figures, exist_ok=True)

    fig, axes = plt.subplots(1, 2, figsize=(12, 5))
    sns.heatmap(Sigma_0, annot=True, fmt=".3f", cmap="RdBu", xticklabels=param_names, yticklabels=param_names, ax=axes[0])
    axes[0].set_title("Prior Covariance Matrix")
    sns.heatmap(Sigma_post, annot=True, fmt=".4f", cmap="PiYG", xticklabels=param_names, yticklabels=param_names, ax=axes[1])
    axes[1].set_title("Posterior Covariance Matrix")

    fig.tight_layout()
    plt.suptitle(f'Posterior Covariance Matrix')
    plt.savefig(f'{output_path_figures}/posterior_covariance_matrix.png') 


def plot_posterior_simulations(dummy_data_dir, output_dir_bayesian):

    true_waveforms = pd.read_csv(f"{dummy_data_dir}/output_dummy_data/waveform_resampled_all_pressure_traces_rv_with_pca.csv")
    posterior_waveforms = pd.read_csv(f"{output_dir_bayesian}/waveform_resampled_all_pressure_traces_rv.csv")
    
    # Ground truth waveform
    which_obs = 3
    y_true = pd.Series(true_waveforms.iloc[which_obs, :101].values)

    # Posterior waveforms
    samples = posterior_waveforms.iloc[:, :101].values # Shape (100,101)
    

    # Compute and plot the mean waveform
    mean_waveform = samples.mean(axis=0)
    var_waveform = samples.var(axis=0) + 1e-6  # Adding a small constant to avoid division by zero
    
    
    fig, ax = plt.subplots(figsize=(10, 5))
    output_path_figures = os.path.join(output_dir_bayesian, "figures")
    os.makedirs(output_path_figures, exist_ok=True)

    
    
    # Inputs
    y_obs = y_true.values                    # shape: (101,)
    posterior_preds = samples               # shape: (100, 101)
    S, T = posterior_preds.shape            # S = 100, T = 101

    # Set Gaussian likelihood standard deviation (fixed)
    sigma = 1 # Dynamically adjust from observation model? Or is it always 1 as it is only concerned with waveform even if calibrated on something else?
    

    # Compute log pointwise predictive density
    log_likelihoods = -0.5 * np.log(2 * np.pi * sigma**2) \
                    - ((y_obs - posterior_preds)**2) / (2 * sigma**2)
    

    
    # log_likelihoods shape: (100, 101)
    # Average over posterior samples (axis 0), then sum over timepoints
    lppd =  np.sum(np.log(np.mean(np.exp(log_likelihoods), axis=0)))
    nlpd = -lppd

    # p_WAIC: sum of variances of log-likelihoods across posterior samples
    p_waic = np.sum(np.var(log_likelihoods, axis=0, ddof=1))  # scalar

    # WAIC computation
    waic = -2 * (lppd - p_waic)

    # Compute RMSE
    sqe = (y_obs - mean_waveform) ** 2
    rmse = np.sqrt(sqe.mean(axis=0))
    
    
    
    # Plot all waveforms in faded orange
    for j in range(samples.shape[0]):
        if j == 0:
            ax.plot(samples[j, :], color='bisque', label="Posterior Samples")
        else:
            ax.plot(samples[j, :], color='bisque', alpha=0.05)
    
    # Plot y_true
    ax.plot(y_true.values, label="True Waveform", color='c', linewidth=2)

    # Plot mean waveform
    ax.plot(mean_waveform, color='darkorange', linewidth=1.5, label="Mean Calibrated Waveform")
    
    
    ax.set_xticks(np.arange(0, 110, 10))
    ax.set_ylim(0,70)
    ax.tick_params(axis='x', labelsize=14)
    ax.tick_params(axis='y', labelsize=14)
    ax.set_xlabel("Time Index", fontsize=16)
    ax.set_title(f"RMSE = {rmse:.4f}, NLPD = {nlpd:.2f}")
    ax.set_ylabel("Pressure (mmHg)", fontsize=16)
    ax.set_yticks(np.arange(0,70, 10))
    
    #ax.legend()

    #fig.suptitle("Calibrated Pressure Waveforms for Different Methods")
    fig.tight_layout()
    fig.show()
    fig.savefig(os.path.join(output_path_figures, "posterior_simulated_waveforms.png"))


def plot_parameter_trajectories(Sigma_post,
                                posterior_means,
                                bc,
                                output_path):
    
    output_path_figures = os.path.join(output_path,"figures")
    os.makedirs(output_path_figures, exist_ok=True)

    class ResolutionController:
        def __init__(self, window_size):
            self.window_size = window_size

        def downsample(self, data):
            """Downsamples the data by averaging over non-overlapping windows."""
            if data.shape[0] < self.window_size:
                raise ValueError(f"Data has fewer than {self.window_size} time steps!")

            num_windows = data.shape[0] // self.window_size  # Compute number of full windows
            return data[:num_windows * self.window_size].reshape(num_windows, self.window_size, -1).mean(axis=1)

    # Initialize resolution controller
    window_size = 10
    res_controller = ResolutionController(window_size)

    # Define time range before downsampling
    time_range = (1, 105000)  # Specify the indices from the original data

    # Ensure posterior_variances has shape (3888, p)
    posterior_variances_corrected = np.array(Sigma_post).diagonal().reshape(1, -1)  # (1, p)
    posterior_variances_corrected = np.tile(posterior_variances_corrected, (posterior_means.shape[0], 1))  # (3888, p)


    # Slice the original data before downsampling
    posterior_means_trimmed = posterior_means[time_range[0]:time_range[1]]
    posterior_variances_trimmed = posterior_variances_corrected[time_range[0]:time_range[1]]

    # Downsample the sliced data
    posterior_means_smooth = res_controller.downsample(posterior_means_trimmed)  # (new_length, p)
    posterior_variances_smooth = res_controller.downsample(np.sqrt(posterior_variances_trimmed))  # (new_length, p)


    # Generate new time indices based on downsampling
    T_smooth = np.arange(posterior_means_smooth.shape[0]) * window_size + time_range[0]

    # Colors for different parameters
    param_names = bc.param_names
    colors = plt.cm.get_cmap('Set1', len(param_names)).colors

    # Plot each parameter on a separate subplot
    fig, axes = plt.subplots(len(param_names), 1, figsize=(10, 8), sharex=True)

    for i in range(len(param_names)):
        mean = posterior_means_smooth[:, i]  # Smoothed mean
        std_dev = posterior_variances_smooth[:, i]  # Smoothed standard deviation

        axes[i].plot(T_smooth, mean, color=colors[i], label=param_names[i])
        axes[i].fill_between(T_smooth, mean - 2 * std_dev, mean + 2 * std_dev, color=colors[i], alpha=0.2)

        axes[i].set_ylabel('Value')
        axes[i].legend(loc='center left', bbox_to_anchor=(1.0, 0.5))
        axes[i].grid()

    axes[-1].set_xlabel('Beats')
    fig.suptitle(f'Parameter Trajectories (Averaged Over {window_size} Steps)')

    plt.tight_layout()
    plt.subplots_adjust(right=0.85)  # Make space for legends on the right
    fig.savefig(os.path.join(output_path_figures, "posterior_simulated_waveforms.png"))


def plot_sensitivity_heatmap(directory, saveto, selected_keys=[]):
    """Plots a heatmap of sensitivity indices for each parameter across all CSV files."""

    output_path_figures = os.path.join(directory, "figures/sensititvity_heatmaps")
    os.makedirs(output_path_figures, exist_ok=True)

    csv_files = [f for f in os.listdir(directory) if f.endswith(".csv")]

    # Load data from selected CSV files.
    files_to_read = csv_files if not selected_keys else selected_keys
    data = {}
    for file in files_to_read:
        file_path = os.path.join(directory, file)
        data[file] = pd.read_csv(file_path, index_col=0)

    combined_df = pd.DataFrame()
    for file_name, df in data.items():
        combined_df[file_name] = df["ST"]

    combined_df = combined_df.fillna(0).T  # Transpose to have CSV files on Y-axis and parameters on X-axis
    combined_df.index = combined_df.index.str.replace('sensitivity_', '', regex=False).str.replace('.csv', '', regex=False)

    # Add a column for row means (mean ST for each output)
    combined_df['Threshold Value'] = combined_df.mean(axis=1)

    # Order columns by the max across output for each parameter (column)
    parameter_max = combined_df.max()
    ordered_columns = parameter_max.sort_values(ascending=False).index.tolist()
    combined_df = combined_df[ordered_columns]  # Reorder columns

    # Set row heights larger by increasing the figure height
    row_height = 1  # Height per row in inches (increase for larger rows)
    fig_height = max(6, row_height * len(combined_df.index))
    plt.figure(figsize=(35, fig_height))

    sns.heatmap(
        combined_df,
        cmap="Greens",
        linewidths=1,
        cbar=False,
        cbar_kws={
            "orientation": "horizontal",
            "shrink": 0.7,
            "pad": 0.3,
            "label": "Sensitivity Index (ST)"
        }
    )

    plt.title("")
    plt.ylabel("Output", fontsize=28, fontweight='bold')
    plt.xticks(rotation=45, fontsize=24)
    plt.xlabel("Parameters", fontsize=28, fontweight='bold')
    plt.yticks(rotation=0, fontsize=24)
    plt.tight_layout()
    plt.savefig(f"{output_path_figures}/{saveto}_sensitivity_heatmap.png", dpi=600)
    plt.savefig(f"{output_path_figures}/{saveto}_sensitivity_heatmap.pdf", dpi=600)

def plot_kf_estimates(estimates, 
                      param_names, 
                      confidence_level=0.95, 
                      figsize=(10, 0.7), 
                      cmap='tab10',
                      output_path=None):
    """
    Plots Kalman filter parameter estimates with confidence intervals, stacked vertically in one figure.
    
    Parameters:
    - estimates: List of (mu_t, Sigma_t) tuples from the KF run.
    - param_names: List of parameter names.
    - confidence_level: Confidence level for intervals (default=0.95).
    - figsize: Tuple specifying width and height per subplot row.
    - cmap: Matplotlib colormap name for assigning colours.
    - output_path: directory to save the plots. A subdirectory will be created called "figures"
    """

    output_path_figures = os.path.join(output_path,"figures")
    os.makedirs(output_path_figures, exist_ok=True)

    n_timesteps = len(estimates)
    n_params = len(param_names)

    z = stats.norm.ppf(0.5 + confidence_level / 2)

    mu_series = np.array([mu for mu, _ in estimates])  # shape: (n_timesteps, n_params)
    sigma_series = np.array([np.sqrt(np.diag(Sigma)) for _, Sigma in estimates])  # shape: (n_timesteps, n_params)

    time = np.arange(n_timesteps)

    fig, axes = plt.subplots(n_params, 1, figsize=(figsize[0], figsize[1]*n_params), sharex=True)

    if n_params == 1:
        axes = [axes]  # ensure axes is iterable

    # Generate distinct colours using a colormap
    color_map = cm.get_cmap(cmap, n_params)
    colors = [color_map(i) for i in range(n_params)]

    for i, ax in enumerate(axes):
        color = colors[i]
        ax.plot(time, mu_series[:, i], color=color, label=param_names[i], linewidth=1)
        ax.fill_between(time, 
                        mu_series[:, i] - z * sigma_series[:, i],
                        mu_series[:, i] + z * sigma_series[:, i],
                        color=color, alpha=0.2)
        
        ax.grid(True, linestyle="--", alpha=0.5)
        ax.legend(loc='upper right', fontsize='small')
    
    axes[-1].set_xlabel("Time Step")

    plt.tight_layout()
    plt.subplots_adjust(right=0.85)

    # Save the figure
    fig.savefig(f'{output_path_figures}/kf_parameter_estimates.png', dpi=300)




import pytest
import pandas as pd
import os
import numpy as np
from cvdnet_pipeline.compute_pca import compute_pca
import shutil


# Parametrize data types for testing both synthetic and real data
DATA_TYPES = ["synthetic", "real"]


@pytest.fixture(params=DATA_TYPES)
def data_type(request):
    """Fixture that provides both synthetic and real data types for parameterized tests."""
    return request.param


@pytest.fixture
def cleanup_output(data_type):
    """Cleanup fixture that handles output files for both data types."""
    if data_type == "synthetic":
        output_file           = f"tests/inputs_for_tests/compute_pca_module/output_64_9_params/{data_type}_data/output_64_9_params/waveform_resampled_all_pressure_traces_rv_with_pca.csv"
        output_pca_folder     = f"tests/inputs_for_tests/compute_pca_module/output_64_9_params/{data_type}_data/output_64_9_params/pca"
        output_figures_folder = f"tests/inputs_for_tests/compute_pca_module/output_64_9_params/{data_type}_data/output_64_9_params/figures"

    elif data_type == "real":
        output_file           = f"tests/inputs_for_tests/compute_pca_module/output_64_9_params/{data_type}_data/waveform_resampled_all_pressure_traces_rv_with_pca.csv"
        output_pca_folder     = f"tests/inputs_for_tests/compute_pca_module/output_64_9_params/{data_type}_data/pca"
        output_figures_folder = f"tests/inputs_for_tests/compute_pca_module/output_64_9_params/{data_type}_data/figures"


    yield output_file, data_type
    
    # Cleanup files and directories
    if os.path.exists(output_file):
        os.remove(output_file)
    if os.path.exists(output_pca_folder):
        shutil.rmtree(output_pca_folder)
    if os.path.exists(output_figures_folder):
        shutil.rmtree(output_figures_folder)                


def test_compute_pca(cleanup_output):
    """Test compute_pca with valid input for both synthetic and real data."""
    output_file, data_type = cleanup_output

    # Call the function with the input file
    compute_pca(n_samples=64, 
                n_params=9,
                n_pca_components=10,
                output_path=f'tests/inputs_for_tests/compute_pca_module/output_64_9_params/{data_type}_data',
                data_type=data_type)

    # Check if the output data matches the expected output
    output_data = pd.read_csv(output_file)
    expected_output = pd.read_csv(f'tests/expected_outputs/compute_pca_module/output_64_9_params/{data_type}_data/waveform_resampled_all_pressure_traces_rv_with_pca.csv')
    pd.testing.assert_frame_equal(output_data, expected_output)


import os
import numpy as np
import pandas as pd
from SALib import ProblemSpec
from multiprocessing import Pool
from functools import partial

def process_emulator(emulator_name,
                     pure_input_params,
                     emulators,
                     output_dir,
                     seed):
    """Process a single emulator for sensitivity analysis"""
    print(f"Processing {emulator_name}...")

    # Define problem spec for sensitivity analysis
    problem = ProblemSpec({
        'num_vars': len(pure_input_params.columns),
        'names': pure_input_params.columns.tolist(),
        'bounds': pure_input_params.describe().loc[['min', 'max']].T.values,
        "outputs": [emulator_name],
    })

    linear_model = emulators.loc[emulator_name, 'Model']

    # Sample inputs
    problem.sample_sobol(1024, seed=seed)
    X_samples = problem.samples
    
    # Extract emulator coefficients and intercept
    beta_matrix = np.array(linear_model.coef_)
    intercept = np.array(linear_model.intercept_)

    # Ensure correct shape
    if beta_matrix.ndim == 1:
        beta_matrix = beta_matrix.reshape(1, -1)
    if intercept.ndim == 0:
        intercept = np.array([intercept])
    
    # Compute emulator outputs
    Y_samples = X_samples @ beta_matrix.T + intercept
    Y_reshape = Y_samples.reshape(-1)

    # Set and analyze results
    problem.set_results(Y_reshape)
    sobol_indices = problem.analyze_sobol(print_to_console=False,
                                          seed=seed)

    # Sort results
    total, first, second = sobol_indices.to_df()
    total.sort_values('ST', inplace=True, ascending=False)

    # Save results
    result_data = pd.DataFrame({
        "Parameter": total.index,
        "ST": total['ST'],
        "ST_conf": total['ST_conf']
    })
    save_emulator_name = emulator_name.replace("/", "_")
    result_file = os.path.join(output_dir, f"sensitivity_{save_emulator_name}.csv")
    result_data.to_csv(result_file, index=False)
    
    return f"Completed {emulator_name}"    


def sensitivity_analysis(n_samples: int, 
                         n_params: int, 
                         output_path: str,
                         seed: int = 42,
                         n_processes: int = None):

    file_suffix = f'_{n_samples}_{n_params}_params'

    # Read Input Data
    pure_input_params = pd.read_csv(f"{output_path}/pure_input{file_suffix}.csv")

    # Import Emulators
    emulators = pd.read_pickle(f"{output_path}/output{file_suffix}/emulators/linear_models_and_r2_scores_{n_samples}.pkl")

    # Extract the emulator names
    emulator_list = emulators.index.to_list()

    # Directory to save results
    output_dir = f"{output_path}/output{file_suffix}/sensitivity_analysis_results"
    os.makedirs(output_dir, exist_ok=True)

    # Create partial function with fixed arguments
    process_func = partial(
        process_emulator,
        pure_input_params=pure_input_params,
        emulators=emulators,
        output_dir=output_dir,
        seed=seed
    )

    # Use multiprocessing
    with Pool(processes=n_processes) as pool:
        results = pool.map(process_func, emulator_list)

    for result in results:
        print(result)   

    print("All analyses completed.")


import os
import pytest
import pandas as pd
from cvdnet_pipeline.simulate_data import simulate_data
import tempfile
import shutil

RTOL_TOLERANCE = 1e-2

def test_simulate_data():
    # Define test parameters

    with tempfile.TemporaryDirectory() as tmp_path:
        print(f"Temporary directory created at: {tmp_path}")

        param_path = os.path.join('./tests/inputs_for_tests/simulate_data_module/', 
                                  "parameters_pulmonary_sensitive_summarystats.json")  # Ensure this file exists with valid parameters
        n_samples = 64
        repeat_simulations = True

        # Call the function
        simulate_data(
            param_path=param_path,
            n_samples=n_samples,
            output_path=str(tmp_path),
            repeat_simulations=repeat_simulations
        )

        # Verify that the output files are created
        input_file = os.path.join(tmp_path,f'input_{n_samples}_9_params.csv')
        output_dir_sims = os.path.join(tmp_path,f'output_{n_samples}_9_params')
        bool_indices_file = os.path.join(output_dir_sims,f'bool_indices_{n_samples}.csv')
        output_dir_pressure_traces_pat = os.path.join(output_dir_sims,'pressure_traces_pat','all_pressure_traces.csv')
        output_dir_pressure_traces_rv = os.path.join(output_dir_sims,'pressure_traces_rv','all_pressure_traces.csv')


        assert os.path.exists(input_file), "Input file was not created."
        assert os.path.exists(output_dir_sims), "Simulations Output directory was not created."
        assert os.path.exists(bool_indices_file), "Bool indices file was not created."


        # Optionally, check the contents of the input file
        input_data = pd.read_csv(input_file)
        assert len(input_data) == n_samples, "Input file does not contain the expected number of samples."

        # Compare the input file to the input file in the expected_outputs directory
        expected_input_file_path = os.path.join('./tests/expected_outputs/simulate_data_module',
                                           f'output_{n_samples}_9_params/',
                                           f'input_{n_samples}_9_params.csv')
        expected_input_data = pd.read_csv(expected_input_file_path)
        pd.testing.assert_frame_equal(input_data, expected_input_data)

        # Compare the output files to the expected output files
        expected_output_dir = os.path.join('./tests/expected_outputs/simulate_data_module',
                                            f'output_{n_samples}_9_params/')
        expected_pressure_traces_pat = pd.read_csv(os.path.join(expected_output_dir,
                                                    'pressure_traces_pat',
                                                    'all_pressure_traces.csv'))
        expected_pressure_traces_rv = pd.read_csv(os.path.join(expected_output_dir,
                                                    'pressure_traces_rv',
                                                    'all_pressure_traces.csv'))

        resulting_pressure_traces_pat = pd.read_csv(output_dir_pressure_traces_pat)
        resulting_pressure_traces_rv = pd.read_csv(output_dir_pressure_traces_rv)

        pd.testing.assert_frame_equal(resulting_pressure_traces_pat, 
                                      expected_pressure_traces_pat,
                                      check_exact=False,
                                      rtol=RTOL_TOLERANCE)
        pd.testing.assert_frame_equal(resulting_pressure_traces_rv, 
                                      expected_pressure_traces_rv,
                                      check_exact=False,
                                      rtol=RTOL_TOLERANCE)


        # delete files to check loading simulations from disk
        os.remove(input_file)
        os.remove(bool_indices_file)
        os.remove(output_dir_pressure_traces_pat)
        os.remove(output_dir_pressure_traces_rv)

        simulate_data(
            param_path=param_path,
            n_samples=n_samples,
            output_path=str(tmp_path),
            repeat_simulations=False
        )
        # Check if the output directory is empty

        assert os.path.exists(input_file), "Input file was not created."
        assert os.path.exists(bool_indices_file), "Bool indices file was not created."
        assert os.path.exists(output_dir_pressure_traces_pat), "PAT pressure traces file was not created."
        assert os.path.exists(output_dir_pressure_traces_rv), "RV pressure traces file was not created."

        # Run the test for calibrated parameters
        output_dir_bayesian = os.path.join('./tests/expected_outputs/calibrate_parameters_module',
                                           'synthetic_data/output_64_9_params/bayesian_calibration_results',
                                           '17_output_keys/calibration_20250827_151335')

        output_dir_sims, n_params = simulate_data(
            param_path=param_path,
            n_samples=n_samples,
            output_path=output_dir_bayesian,
            sample_parameters = False
        )

        # Check that the 'posterior_simulations' folder is created
        assert os.path.exists(os.path.join(output_dir_bayesian, 'posterior_simulations')), "Posterior simulations directory was not created."

        # Compare the output files to the expected output files
        expected_output_dir = os.path.join('./tests/expected_outputs/simulate_data_module',
                                            f'output_{n_samples}_9_params/',
                                            'bayesian_calibration_results/17_output_keys/calibration_20250827_151335/')
        expected_pressure_traces_pat = pd.read_csv(os.path.join(expected_output_dir,
                                                    'pressure_traces_pat',
                                                    'all_pressure_traces.csv'))
        expected_pressure_traces_rv = pd.read_csv(os.path.join(expected_output_dir,
                                                    'pressure_traces_rv',
                                                    'all_pressure_traces.csv'))

        resulting_pressure_traces_pat = pd.read_csv(os.path.join(output_dir_bayesian,'pressure_traces_pat','all_pressure_traces.csv'))
        resulting_pressure_traces_rv = pd.read_csv(os.path.join(output_dir_bayesian,'pressure_traces_rv','all_pressure_traces.csv'))

        pd.testing.assert_frame_equal(resulting_pressure_traces_pat, 
                                      expected_pressure_traces_pat,
                                      check_exact=False,
                                      rtol=RTOL_TOLERANCE)
        pd.testing.assert_frame_equal(resulting_pressure_traces_rv, 
                                      expected_pressure_traces_rv,
                                      check_exact=False,
                                      rtol=RTOL_TOLERANCE)

        # Delete the output directory to clean up
        shutil.rmtree(os.path.join(output_dir_bayesian,'figures'))
        shutil.rmtree(os.path.join(output_dir_bayesian,'pressure_traces_pat'))
        shutil.rmtree(os.path.join(output_dir_bayesian,'pressure_traces_rv'))
        shutil.rmtree(os.path.join(output_dir_bayesian,'posterior_simulations'))        
        os.remove(os.path.join(output_dir_bayesian, 'bool_indices_64.csv'))



"""
Constants and configurations used across the pipeline.
"""

VALID_PIPELINE_STEPS = ["sim", # simulation
                  "ag", # analyse giessen
                  "pca", # principal component analysis
                  "emu", # building emulations
                  "cal", # calibration
                  "kf", # Kalman filter
                  "post_sim", # posterior simulation
                  "post_res"] # posterior resampling


# Import necessary libraries
import numpy as np

class KalmanFilterWithEmulator:
    def __init__(self, B, B_0, Q, R, Sigma_emu, mu_0, Sigma_0):
        """
        Parameters:
        - B: (n_obs x n_params) emulator linear coefficient matrix
        - B_0: (n_obs,) emulator intercept vector
        - Q: (n_params x n_params) state transition covariance (process noise)
        - R: (n_obs x n_obs) observation noise covariance
        - Sigma_emu: (n_obs x n_obs) emulator uncertainty covariance (diagonal from emulator RSEs)
        - mu_0: (n_params,) prior mean of theta
        - Sigma_0: (n_params x n_params) prior covariance of theta
        """
        self.B = B
        self.B_0 = B_0
        self.Q = Q
        self.R = R
        self.Sigma_emu = Sigma_emu
        self.Sigma_obs_total = Sigma_emu + R

        self.mu = mu_0
        self.Sigma = Sigma_0


    def step(self, y_t):
        """
        Perform one Kalman update step given observation y_t.
        Returns posterior mean and covariance of theta at this time.
        """
        # Prediction : State transition matrix F is implicitly the identity i.e. a random walk
        mu_pred = self.mu
        Sigma_pred = self.Sigma + self.Q

        # Kalman gain
        S = self.B @ Sigma_pred @ self.B.T + self.Sigma_obs_total
        K = Sigma_pred @ self.B.T @ np.linalg.inv(S)

        # Innovation (residual)
        innovation = y_t - (self.B @ mu_pred + self.B_0)

        # Update
        self.mu = mu_pred + K @ innovation
        self.Sigma = (np.eye(len(self.mu)) - K @ self.B) @ Sigma_pred

        return self.mu, self.Sigma


    def run(self, Y):
        """
        Run the filter on a sequence of observations.
        Y: (n_timesteps x n_obs) array of observations
        Returns: list of (mu_t, Sigma_t) at each time step
        """
        estimates = []
        for y_t in Y:
            mu_t, Sigma_t = self.step(y_t)
            estimates.append((mu_t.copy(), Sigma_t.copy()))
        return estimates


import pytest
import pandas as pd
import os
import numpy as np
from cvdnet_pipeline.build_emulator import build_emulator
import shutil

@pytest.fixture
def cleanup_output():
    output_file = "tests/inputs_for_tests/build_emulator_module/output_50_9_params/emulators/linear_models_and_r2_scores_50.csv"

    yield output_file
    if os.path.exists(output_file):
        shutil.rmtree('tests/inputs_for_tests/build_emulator_module/output_50_9_params/emulators/')

def test_build_emulator(cleanup_output):


    output_keys = [
        "t_max_dpdt",
        "a_epad",
        "epad",
        "s_a_epad",
        "s_epad",
        "A_p",
        "P_max",
        "esp",
        "sys",
        "iT",
        "PC1",
        "PC2",
        "PC3"
        ]

    # Call the function with the input file
    build_emulator(n_samples=50, 
                   n_params=9,
                   output_path='tests/inputs_for_tests/build_emulator_module',
                   output_file_name="waveform_resampled_all_pressure_traces_rv_with_pca.csv",
                   output_keys_red=output_keys)

    # Check if the output data matches the expected output
    output_data = pd.read_csv(cleanup_output)
    expected_output = pd.read_csv('tests/expected_outputs/build_emulator_module/output_50_9_params/emulators/linear_models_and_r2_scores_50.csv')
    pd.testing.assert_frame_equal(output_data, expected_output)


# utils/io.py
import numpy as np
import pandas as pd
import os
from sklearn.linear_model import LinearRegression
from sklearn.metrics import r2_score
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split

def load_csv(path, drop_column=""):
    return pd.read_csv(path, usecols=lambda x: x != drop_column)

def save_csv(df, path):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    df.to_csv(path, index=False)

def load_simulation(input_path):
    simulations = []
    for file in os.listdir(input_path):
        if file.startswith('all_outputs') and file.endswith('.csv'):
            df = pd.read_csv(os.path.join(input_path, file), index_col=0)
            simulations.append(df)
    return simulations

def select_feasible_traces(simulated_traces, output_path):
    # Create column headers
    headers = list(range(100)) + ['CO', 'dt', 'EF', 'dPAP', 'sPAP', 'mPAP']

    # List to collect all pressure traces
    pressure_traces_list_pat = []
    pressure_traces_list_rv = []

    for ind in range(len(simulated_traces)):
        if not isinstance(simulated_traces[ind], bool):

            # PAT pressure
            p_pat_raw = simulated_traces[ind]['p_pat'].values.copy()

            # RV pressure
            p_rv_raw = simulated_traces[ind]['p_rv'].values.copy()

            T = simulated_traces[ind]['T'].values.copy()
            T_resample = np.linspace(T[0], T[-1], 100)

            # Interpolate pressure for 100 timesteps from 1000
            p_pat_resampled = np.interp(T_resample, T, p_pat_raw)
            p_rv_resampled = np.interp(T_resample, T, p_rv_raw)

            # Compute CO
            q_pat = simulated_traces[ind]['q_pat'].values.copy()
            CO = np.sum(q_pat) * (T[1] - T[0]) / (T[-1] - T[0]) * 60. / 1000.  # L / min

            # Compute EF
            v_rv = simulated_traces[ind]['v_rv'].values.copy()
            EF = (np.max(v_rv) - np.min(v_rv)) / np.max(v_rv)

            # Compute dPAP, sPAP, mPAP
            dPAP = min(p_rv_raw)
            sPAP = max(p_rv_raw)
            mPAP = np.mean(p_rv_raw)

            # Record time interval, approx T (input param) / 100, there are some rounding differences due to interpolation
            tl = T_resample - simulated_traces[ind]['T'].iloc[0]
            dt = np.diff(tl)[0]

            # Only create array if conditions hold or screening is turned off
            #if (2 < CO < 12 and 4 < dPAP < 67 and 9 < mPAP < 87 and 15 < sPAP < 140):
                # Create a 2D array for saving
            pressure_trace_pat = np.hstack((p_pat_resampled, [CO], [dt], [EF], [dPAP], [sPAP], [mPAP]))
            pressure_trace_rv = np.hstack((p_rv_resampled, [CO], [dt], [EF], [dPAP], [sPAP], [mPAP]))
            pressure_traces_list_pat.append(pressure_trace_pat)
            pressure_traces_list_rv.append(pressure_trace_rv)

    # Convert the list of pressure traces to a DataFrame
    pressure_traces_df_pat = pd.DataFrame(pressure_traces_list_pat, columns=headers)
    pressure_traces_df_rv = pd.DataFrame(pressure_traces_list_rv, columns=headers)

    return pressure_traces_df_pat, pressure_traces_df_rv

def emulate_linear(input, output):
    # Input and output data
    X = input
    Y = output

    # Initialize the model
    model = LinearRegression()
    

    # Fit the model to the training data
    model.fit(X, Y)

    # Predict the output for the test data
    y_pred = model.predict(X)

    # Compute R² score for the predictions versus actual test data
    r2 = r2_score(Y, y_pred)

    # compute MSE
    mse = mean_squared_error(Y, y_pred)

    # compute RMSE
    rmse = np.sqrt(mse)

    # compute RSE
    n = len(Y)   # Number of observations
    p = X.shape[1] + 1  # Number of parameters (including intercept)
    rss = np.sum((Y - y_pred) ** 2)  # Residual Sum of Squares
    rse = np.sqrt(rss / (n - p))


    return model, r2, mse, rmse, rse


import os
import pandas as pd
from cvdnet_pipeline.utils import helper_functions


# steps/build_emulator.py
def build_emulator(n_samples:int=200, 
                   n_params:int=9, 
                   output_path:str="output", 
                   output_file_name:str="waveform_resampled_all_pressure_traces_rv_with_pca.csv",
                   output_keys_red:list[str]=None):
    
    file_suffix = f'_{n_samples}_{n_params}_params'

    input_file = pd.read_csv(f"{output_path}/pure_input{file_suffix}.csv")
    output_file = pd.read_csv(f"{output_path}/output{file_suffix}/{output_file_name}")

    # Select only first relevant inputs 
    filtered_input = input_file.copy()

    # List of output keys to process
    output_keys = output_file.columns
    
    # Initialize dictionaries to store R2 scores and models
    linear_r2_scores = {}
    linear_mse_scores = {}
    linear_rmse_scores = {}
    linear_rse_scores = {}
    fitted_models = {}

    # Iterate through the output keys:

    # Here, we take each output, such as the raw pressure traces, calculated clinical indices and the PCA components,
    # (if present) and fit a linear model to each of them using those parameters that have been selected as relevant inputs.
    # Currently, just a linear regression model is used.
    for key in output_keys:
        model, r2, mse, rmse, rse = helper_functions.emulate_linear(input=filtered_input, output=output_file[key])
        linear_r2_scores[key] = r2
        linear_mse_scores[key] = mse
        linear_rmse_scores[key] = rmse
        linear_rse_scores[key] = rse
        fitted_models[key] = model

    # Convert the dictionaries to a DataFrame
    emulator_results_df = pd.DataFrame({'R2_Score': linear_r2_scores, 
                                        'MSE': linear_mse_scores,
                                        'RMSE': linear_rmse_scores,
                                        'RSE': linear_rse_scores, 
                                        'Model': fitted_models})
    
    # Create directory for output if it doesn't exist
    if not os.path.exists(f"{output_path}/output{file_suffix}/emulators"):
        os.makedirs(f"{output_path}/output{file_suffix}/emulators")

    # Save the DataFrame to a CSV file
    emulator_results_df.to_csv(f'{output_path}/output{file_suffix}/emulators/linear_models_and_r2_scores_{n_samples}.csv')

    # To save the DataFrame with models, use pickle
    emulator_results_df.to_pickle(f'{output_path}/output{file_suffix}/emulators/linear_models_and_r2_scores_{n_samples}.pkl')
    
    # Save reduced Dataframe to a csv
    if output_keys_red is not None:
        reduced_emulator_results_df = emulator_results_df.loc[output_keys_red].copy()
        reduced_emulator_results_df.to_csv(f'{output_path}/output{file_suffix}/emulators/calibration_features_linear_models_and_r2_scores_{n_samples}.csv')
    



"""
Constants and configurations used across test modules.
"""

# Output keys for different test scenarios
OUTPUT_KEYS_FOR_TESTS_FOR_REAL_DATA_CALIBRATION = [
    "t_max_dpdt", 
    "a_epad",
    "epad", 
    "s_a_epad", 
    "s_epad",
    "min_dpdt", 
    "max_dpdt", 
    "A_p", 
    "P_max", 
    "esp", 
    "sys",
    "iT", 
    "PC1", 
    "PC2", 
    "PC3"
]

OUTPUT_KEYS_FOR_TESTS_FOR_SYNTHETIC_DATA_CALIBRATION = [
    "t_max_dpdt", 
    "a_epad", 
    "epad", 
    "s_a_epad", 
    "s_epad",
    "min_dpdt", 
    "max_dpdt", 
    "A_p", 
    "P_max", 
    "esp",
    "sys",
    "EF", 
    "Ees/Ea", 
    "iT", 
    "PC1", 
    "PC2", 
    "PC3"
]

# Common test parameters
DEFAULT_N_SAMPLES = 64
DEFAULT_N_PARAMS = 9
DEFAULT_EPSILON_OBS_SCALE = 0.05


import pandas as pd
from giessen_data_analysis import analyseGiessen

def analyse_giessen(file_path: str, data_type: str, gaussian_sigmas : list[float]):

    if data_type == 'synthetic':
        rv_file = pd.read_csv(f"{file_path}/pressure_traces_rv/all_pressure_traces.csv")
        ar_file = pd.read_csv(f"{file_path}/pressure_traces_pat/all_pressure_traces.csv")
        
        # Flag and remove any negative pressure traces
        negative_indices = rv_file[(rv_file < 0).any(axis=1)].index
        if not negative_indices.empty:
            print(f"Removing {len(negative_indices)} negative RV pressure traces at indices: {negative_indices.tolist()}")
            rv_file = rv_file.drop(index=negative_indices).reset_index(drop=True)
        
        # unpack sigmas
        sigma_filter_pressure, sigma_filter_dpdt, sigma_filter_d2pdt2 = gaussian_sigmas

        # Here takes all Pressure traces from step 1

        all_pressure_traces = pd.DataFrame()
        for ind in range(len(rv_file)):
            if ind % 1000 == 0:
                print(f"Processing {ind}th trace")
            dt = rv_file.loc[ind, 'dt']
            f = rv_file.iloc[[ind], :100].T
            f_repeated = pd.concat([f] * 5, axis=0, ignore_index=True)
            f_repeated.columns = ["Pressure"]
            f_repeated["cPressure"] = f_repeated['Pressure']

            ag = analyseGiessen(df=f_repeated, t_resolution=dt)
            ag.sigma_filter_pressure = sigma_filter_pressure
            ag.sigma_filter_dpdt     = sigma_filter_dpdt
            ag.sigma_filter_d2pdt2   = sigma_filter_d2pdt2

            ag.compute_derivatives()
            # Should change start_at_edp bask to False
            ag.compute_points_of_interest(height=10, use_filter=False, start_at_edp=False) # , export_true_derivates=True, export_true_p=True, distance=90 (we should consder adding these options)
            beats = pd.DataFrame(ag.resample_heart_beat())
            sumstats = ag.points_df

            resampled_df = pd.concat([beats, sumstats.iloc[:-1, :]], axis=1)
            # overwrite EF with model version
            resampled_df['MC_EF'] = rv_file.loc[ind, 'EF']
            # add MC version of edp
            resampled_df['MC_edp'] = rv_file.loc[ind, "0"]
            # add MC version of eivc
            f_pat = ar_file.iloc[[ind],:100].T
            min_ind = f_pat.idxmin().values[0]
            resampled_df['MC_eivc'] = rv_file.loc[ind, min_ind]
            # add MC version of dia
            resampled_df['MC_dia'] = rv_file.loc[ind].min()
            # add t_pulse
            resampled_df['t_pulse'] = (100 + resampled_df['dia_ind'] - resampled_df['edp_ind']) * dt
            
            all_pressure_traces = pd.concat([all_pressure_traces, resampled_df.iloc[[2]]], axis=0)

        all_pressure_traces.reset_index(drop=True, inplace=True)
        all_pressure_traces.to_csv(f"{file_path}/waveform_resampled_all_pressure_traces_rv.csv", index=False)

    elif data_type == 'real':
        
        rv_file = pd.read_csv(f"{file_path}/raw_data.csv")

         # Here takes all Pressure traces from step 1
        rv_file[["Pressure", "cPressure"]] = rv_file[["Pressure [mmHg]", "Compensated Pressure [mmHg]"]]

        ag = analyseGiessen(df=rv_file)

        ag.compute_derivatives()
        ag.compute_points_of_interest()
        beats = pd.DataFrame(ag.resample_heart_beat())
        sumstats = ag.points_df

        resampled_df = pd.concat([beats, sumstats.iloc[:-1, :]], axis=1)
        resampled_df.to_csv(f"{file_path}/waveform_resampled_all_pressure_traces_rv.csv", index=False)


import pandas as pd
import os
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler, PowerTransformer
from sklearn.pipeline import Pipeline
from cvdnet_pipeline.utils import plot_utils


def compute_pca(n_samples:int=50, 
                n_params:int=9, 
                n_pca_components:int=10, 
                output_path:str="output", 
                data_type: str="synthetic"):
    
    output_file_name = 'waveform_resampled_all_pressure_traces_rv.csv' 

    if data_type == "synthetic":
           
        file_sufix = f'_{n_samples}_{n_params}_params'

        dir_name = f"{output_path}/output{file_sufix}"

    elif data_type == "real":

        dir_name = output_path

    output_file = pd.read_csv(f"{dir_name}/{output_file_name}")

    # Create directory for results
    if not os.path.exists(f"{dir_name}/pca"):
        os.makedirs(f"{dir_name}/pca")

    # Create directory for figures
    if not os.path.exists(f"{dir_name}/figures"):
        os.makedirs(f"{dir_name}/figures")

    ## Conduct PCA ##
    df = output_file.copy()

    # Copy the data and separate the target variable (only pressure traces)
    X = df.iloc[:,:100].copy() # traces only

    # Create an instance of the pipeline including StandardScaler and PCA
    pipeline = Pipeline(
        [ ('scl', StandardScaler()),
        ('pca', PCA(n_components=n_pca_components)),
        ('post',PowerTransformer())]
    )

    # Fit the pipeline to the data
    X_pca = pipeline.fit_transform(X)

    # Convert to dataframe
    component_names = [f"PC{i+1}" for i in range(X_pca.shape[1])]
    X_pca = pd.DataFrame(X_pca, columns=component_names, index=df.index)

    X_pca.to_csv(f'{dir_name}/pca/PCA.csv', index=False)

    # Concatenate the PCA components with the original data     
    df_pca = pd.concat([df, X_pca], axis=1)

    # Create a new name for the output file which appends "_with_pca" to the original name
    output_file_name_pca = output_file_name.replace('.csv', '_with_pca.csv')

    df_pca.to_csv(f'{dir_name}/{output_file_name_pca}', index=False)

    # Plot the PCA histogram
    plot_utils.plot_pca_histogram(X_pca, output_path=dir_name, n_pca_components=n_pca_components)

    # Plot the explained variance ratio
    plot_utils.plot_pca_explained_variance(pipeline, output_path=dir_name)

    # Plot the PCA transformed data
    plot_utils.plot_pca_transformed(pipeline, X, output_path=dir_name)